@@ -37,18 +37,17 @@
 #include <QWhatsThis>
 
 // BEGIN KateViewSpace
-KateViewSpace::KateViewSpace(KateViewManager *viewManager, QWidget *parent, const char *name)
+KateViewSpace::KateViewSpace(KateViewManager *viewManager, QWidget *parent)
     : QWidget(parent)
     , m_viewManager(viewManager)
     , m_isActiveSpace(false)
 {
-    setObjectName(QString::fromLatin1(name));
-    QVBoxLayout *layout = new QVBoxLayout(this);
+    auto *layout = new QVBoxLayout(this);
     layout->setSpacing(0);
     layout->setContentsMargins(0, 0, 0, 0);
 
     // BEGIN tab bar
-    QHBoxLayout *hLayout = new QHBoxLayout();
+    auto *hLayout = new QHBoxLayout();
     hLayout->setSpacing(0);
     hLayout->setContentsMargins(0, 0, 0, 0);
 
@@ -113,10 +112,11 @@ KateViewSpace::KateViewSpace(KateViewManager *viewManager, QWidget *parent, cons
     KAcceleratorManager::setNoAccel(m_quickOpen);
     hLayout->addWidget(m_quickOpen);
 
+    auto mwActionCollection = m_viewManager->mainWindow()->actionCollection();
     // forward tab bar quick open action to global quick open action
-    QAction *bridge = new QAction(QIcon::fromTheme(QStringLiteral("quickopen")), i18nc("indicator for more documents", "+%1", 100), this);
+    auto *bridge = new QAction(QIcon::fromTheme(QStringLiteral("quickopen")), i18nc("indicator for more documents", "+%1", 100), this);
     m_quickOpen->setDefaultAction(bridge);
-    QAction *quickOpen = m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_quick_open"));
+    QAction *quickOpen = mwActionCollection->action(QStringLiteral("view_quick_open"));
     Q_ASSERT(quickOpen);
     bridge->setToolTip(quickOpen->toolTip());
     bridge->setWhatsThis(i18n("Click here to switch to the Quick Open view."));
@@ -127,13 +127,14 @@ KateViewSpace::KateViewSpace(KateViewManager *viewManager, QWidget *parent, cons
     m_split->setAutoRaise(true);
     m_split->setPopupMode(QToolButton::InstantPopup);
     m_split->setIcon(QIcon::fromTheme(QStringLiteral("view-split-left-right")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_split_vert")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_split_horiz")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_split_vert_move_doc")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_split_horiz_move_doc")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_close_current_space")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_close_others")));
-    m_split->addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_hide_others")));
+    m_split->setMenu(new QMenu(this));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_split_vert")));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_split_horiz")));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_split_vert_move_doc")));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_split_horiz_move_doc")));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_close_current_space")));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_close_others")));
+    m_split->menu()->addAction(mwActionCollection->action(QStringLiteral("view_hide_others")));
     m_split->setToolTip(i18n("Split View"));
     m_split->setWhatsThis(i18n("Control view space splitting"));
 
@@ -150,6 +151,15 @@ KateViewSpace::KateViewSpace(KateViewManager *viewManager, QWidget *parent, cons
     m_split->installEventFilter(this);
     // END tab bar
 
+    // more explicit enabling of the view space, see bug 485210
+    // we need to setMenu above and watch for aboutToShow to get that right
+    connect(m_quickOpen, &QAbstractButton::pressed, this, [this]() {
+        makeActive(true);
+    });
+    connect(m_split->menu(), &QMenu::aboutToShow, this, [this]() {
+        makeActive(true);
+    });
+
     m_urlBar = new KateUrlBar(this);
 
     // like other editors, we try to re-use documents, of not modified
@@ -221,7 +231,7 @@ bool KateViewSpace::eventFilter(QObject *obj, QEvent *event)
     if (auto button = qobject_cast<QToolButton *>(obj)) {
         // quick open button: show tool tip with shortcut
         if (button == m_quickOpen && event->type() == QEvent::ToolTip) {
-            QHelpEvent *e = static_cast<QHelpEvent *>(event);
+            auto *e = static_cast<QHelpEvent *>(event);
             QAction *quickOpen = m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("view_quick_open"));
             Q_ASSERT(quickOpen);
             QToolTip::showText(e->globalPos(),
@@ -232,7 +242,7 @@ bool KateViewSpace::eventFilter(QObject *obj, QEvent *event)
 
         // quick open button: What's This
         if (button == m_quickOpen && event->type() == QEvent::WhatsThis) {
-            QHelpEvent *e = static_cast<QHelpEvent *>(event);
+            auto *e = static_cast<QHelpEvent *>(event);
             const int hiddenDocs = hiddenDocuments();
             QString helpText = (hiddenDocs == 0)
                 ? i18n("Click here to switch to the Quick Open view.")
@@ -244,24 +254,20 @@ bool KateViewSpace::eventFilter(QObject *obj, QEvent *event)
         }
 
         // on mouse press on view space bar tool buttons: activate this space
-        if (button && !isActiveSpace() && event->type() == QEvent::MouseButtonPress) {
-            m_viewManager->setActiveSpace(this);
-            if (currentView()) {
-                m_viewManager->activateView(currentView()->document(), this);
-            }
+        if (event->type() == QEvent::MouseButtonPress) {
+            makeActive(true);
         }
     }
 
     // ensure proper view space activation even for widgets
-    if (event->type() == QEvent::FocusIn && !isActiveSpace() && currentWidget()) {
-        m_viewManager->setActiveSpace(this);
-        activateWidget(currentWidget());
+    if (event->type() == QEvent::FocusIn) {
+        makeActive(true);
         return false;
     }
 
     // keep track of new sub-widgets
     if (event->type() == QEvent::ChildAdded || event->type() == QEvent::ChildRemoved) {
-        QChildEvent *c = static_cast<QChildEvent *>(event);
+        auto *c = static_cast<QChildEvent *>(event);
         if (c->added()) {
             c->child()->installEventFilter(this);
         } else if (c->removed()) {
@@ -362,9 +368,11 @@ KTextEditor::View *KateViewSpace::createView(KTextEditor::Document *doc)
     // restore the config of this view if possible
     if (!m_group.isEmpty()) {
         if (KateSession::Ptr as = KateApp::self()->sessionManager()->activeSession(); as->config()) {
-            // try id, fallback to url for old configs
-            QString vgroup = QStringLiteral("%1 %2").arg(m_group, KateApp::self()->documentManager()->documentInfo(v->document())->sessionConfigId);
+            // try id first
+            QString id = QString::number(KateApp::self()->documentManager()->documentInfo(v->document())->sessionConfigId);
+            QString vgroup = QStringLiteral("%1 %2").arg(m_group, id);
             if (!as->config()->hasGroup(vgroup)) {
+                // fallback to url, see KateViewSpace::saveViewConfig
                 vgroup = QStringLiteral("%1 %2").arg(m_group, v->document()->url().toString());
             }
             if (as->config()->hasGroup(vgroup)) {
@@ -493,9 +501,11 @@ bool KateViewSpace::showView(DocOrWidget docOrWidget)
     if (index < 0) {
         return false;
     }
-    // move view to end of list
-    m_registeredDocuments.removeAt(index);
-    m_registeredDocuments.push_back(docOrWidget);
+    // move view to end of list if not already there
+    if (index != m_registeredDocuments.size() - 1) {
+        m_registeredDocuments.removeAt(index);
+        m_registeredDocuments.push_back(docOrWidget);
+    }
 
     /**
      * show the wanted view
@@ -534,7 +544,7 @@ void KateViewSpace::changeView(int idx)
         m_viewManager->setActiveSpace(this);
     }
 
-    const auto docOrWidget = m_tabBar->tabDocument(idx);
+    const DocOrWidget docOrWidget = m_tabBar->tabDocument(idx);
 
     // tell the view manager to show the view
     m_viewManager->activateView(docOrWidget, this);
@@ -561,8 +571,12 @@ void KateViewSpace::makeActive(bool focusCurrentView)
 {
     if (!isActiveSpace()) {
         m_viewManager->setActiveSpace(this);
-        if (focusCurrentView && currentView()) {
-            m_viewManager->activateView(currentView()->document(), this);
+        if (focusCurrentView) {
+            if (currentView()) {
+                m_viewManager->activateView(currentView()->document(), this);
+            } else if (currentWidget()) {
+                activateWidget(currentWidget());
+            }
         }
     }
     Q_ASSERT(isActiveSpace());
@@ -580,7 +594,7 @@ void KateViewSpace::registerDocument(KTextEditor::Document *doc)
     /**
      * remember our new document
      */
-    m_registeredDocuments.insert(0, doc);
+    m_registeredDocuments.push_back(doc);
 
     /**
      * ensure we cleanup after document is deleted, e.g. we remove the tab bar button
@@ -603,12 +617,9 @@ void KateViewSpace::registerDocument(KTextEditor::Document *doc)
      */
     connect(doc, &KTextEditor::Document::documentNameChanged, this, &KateViewSpace::updateDocumentName);
     connect(doc, &KTextEditor::Document::documentUrlChanged, this, &KateViewSpace::updateDocumentUrl);
-    connect(doc, &KTextEditor::Document::modifiedChanged, this, [this](KTextEditor::Document *doc) {
-        int tab = m_tabBar->documentIdx(doc);
-        if (tab >= 0) {
-            m_tabBar->setModifiedStateIcon(tab, doc);
-        }
-    });
+    connect(doc, &KTextEditor::Document::modifiedChanged, this, &KateViewSpace::updateDocumentIcon);
+    // needed to get mime-type udate right, see bug 489452
+    connect(doc, &KTextEditor::Document::reloaded, this, &KateViewSpace::updateDocumentIcon);
 
     /**
      * allow signals again, now that the tab is there
@@ -697,7 +708,7 @@ void KateViewSpace::dropEvent(QDropEvent *e)
         e->accept();
         return;
     }
-    auto droppedData = TabMimeData::data(e->mimeData());
+    std::optional<TabMimeData::DroppedData> droppedData = TabMimeData::data(e->mimeData());
     if (droppedData.has_value()) {
         auto doc = KateApp::self()->documentManager()->openUrl(droppedData.value().url);
         auto view = m_viewManager->activateView(doc, this);
@@ -778,7 +789,7 @@ void KateViewSpace::documentDestroyed(QObject *doc)
      * WARNING: this pointer is half destroyed
      * only good enough to check pointer value e.g. for hashes
      */
-    KTextEditor::Document *invalidDoc = static_cast<KTextEditor::Document *>(doc);
+    auto *invalidDoc = static_cast<KTextEditor::Document *>(doc);
     if (m_registeredDocuments.removeAll(invalidDoc) == 0) {
         // do nothing if this document wasn't registered for this viewspace
         return;
@@ -810,6 +821,15 @@ void KateViewSpace::updateDocumentName(KTextEditor::Document *doc)
     }
 }
 
+void KateViewSpace::updateDocumentIcon(KTextEditor::Document *doc)
+{
+    // update tab button if available, might not be the case for tab limit set!
+    const int buttonId = m_tabBar->documentIdx(doc);
+    if (buttonId >= 0) {
+        m_tabBar->setModifiedStateIcon(buttonId, doc);
+    }
+}
+
 void KateViewSpace::updateDocumentUrl(KTextEditor::Document *doc)
 {
     // update tab button if available, might not be the case for tab limit set!
@@ -984,6 +1004,7 @@ bool KateViewSpace::activateWidget(QWidget *widget)
     m_registeredDocuments.removeOne(widget);
     m_registeredDocuments.push_back(widget);
     m_tabBar->setCurrentDocument(DocOrWidget(widget));
+    widget->setFocus();
     return true;
 }
 
@@ -1004,6 +1025,17 @@ void KateViewSpace::setScrollSyncToolButtonVisible(bool visible)
     m_scrollSync->setHidden(!visible);
 }
 
+void KateViewSpace::detachDocument(KTextEditor::Document *doc)
+{
+    auto mainWindow = viewManager()->mainWindow()->newWindow();
+    mainWindow->viewManager()->openViewForDoc(doc);
+
+    // use single shot as this action can trigger deletion of this viewspace!
+    QTimer::singleShot(0, this, [this, doc]() {
+        closeDocument(doc);
+    });
+}
+
 void KateViewSpace::addPositionToHistory(const QUrl &url, KTextEditor::Cursor c, bool calledExternally)
 {
     // We don't care about invalid urls (Fixed Diff View / Untitled docs)
@@ -1117,6 +1149,7 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
     QAction *aCloseTab = menu.addAction(QIcon::fromTheme(QStringLiteral("tab-close")), i18n("&Close Document"));
     QAction *aCloseOthers = menu.addAction(QIcon::fromTheme(QStringLiteral("tab-close-other")), i18n("Close Other &Documents"));
     QAction *aCloseAll = menu.addAction(QIcon::fromTheme(QStringLiteral("tab-close-all")), i18n("Close &All Documents"));
+    menu.addAction(m_viewManager->mainWindow()->actionCollection()->action(QStringLiteral("reopen_latest_closed_document")));
     menu.addSeparator();
     menu.addAction(KStandardAction::open(m_viewManager, &KateViewManager::slotDocumentOpen, this));
     QAction *aDetachTab = menu.addAction(QIcon::fromTheme(QStringLiteral("tab-detach")), i18n("D&etach Document"));
@@ -1124,6 +1157,7 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
     aDetachTab->setWhatsThis(i18n("Opens the document in a new window and closes it in the current window"));
     menu.addSeparator();
     QAction *aCopyPath = addActionFromCollection(&menu, "file_copy_filepath");
+    QAction *aCopyFilename = addActionFromCollection(&menu, "file_copy_filename");
     QAction *aOpenFolder = addActionFromCollection(&menu, "file_open_containing_folder");
     QAction *aFileProperties = addActionFromCollection(&menu, "file_properties");
     menu.addSeparator();
@@ -1136,7 +1170,7 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
         DiffWidgetManager::diffDocs(activeDocument, doc, m_viewManager->mainWindow()->wrapper());
     });
 
-    QMenu *compareUsing = new QMenu(i18n("Compare with Active Document Using"), &menu);
+    auto *compareUsing = new QMenu(i18n("Compare with Active Document Using"), &menu);
     compareUsing->setIcon(QIcon::fromTheme(QStringLiteral("vcs-diff")));
     menu.addMenu(compareUsing);
 
@@ -1148,6 +1182,7 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
 
     if (doc->url().isEmpty()) {
         aCopyPath->setEnabled(false);
+        aCopyFilename->setEnabled(false);
         aOpenFolder->setEnabled(false);
         aRenameFile->setEnabled(false);
         aDeleteFile->setEnabled(false);
@@ -1162,8 +1197,11 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
     }
 
     if (compareUsing->isEnabled()) {
-        for (auto &&diffTool : KateFileActions::supportedDiffTools()) {
+        for (std::pair<QString, QString> &diffTool : KateFileActions::supportedDiffTools()) {
             QAction *compareAction = compareUsing->addAction(diffTool.first);
+            connect(compareAction, &QAction::triggered, this, [tool = diffTool.second, doc, activeDocument] {
+                KateFileActions::compareWithExternalProgram(activeDocument, doc, tool);
+            });
 
             // we use the full path to safely execute the tool, disable action if no full path => tool not found
             compareAction->setData(diffTool.second);
@@ -1191,6 +1229,8 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
         });
     } else if (choice == aCopyPath) {
         KateFileActions::copyFilePathToClipboard(doc);
+    } else if (choice == aCopyFilename) {
+        KateFileActions::copyFileNameToClipboard(doc);
     } else if (choice == aOpenFolder) {
         KateFileActions::openContainingFolder(doc);
     } else if (choice == aFileProperties) {
@@ -1203,50 +1243,58 @@ void KateViewSpace::showContextMenu(int idx, const QPoint &globalPos)
             KateFileActions::deleteDocumentFile(this, doc);
         });
     } else if (choice->parent() == compareUsing) {
-        QString actionData = choice->data().toString(); // name of the executable of the diff program
-        if (!KateFileActions::compareWithExternalProgram(activeDocument, doc, actionData)) {
-            QMessageBox::information(this,
-                                     i18n("Could not start program"),
-                                     i18n("The selected program could not be started. Maybe it is not installed."),
-                                     QMessageBox::StandardButton::Ok);
-        }
+        const QString actionData = choice->data().toString(); // name of the executable of the diff program
+        KateFileActions::compareWithExternalProgram(activeDocument, doc, actionData);
     } else if (choice == aDetachTab) {
-        auto mainWindow = viewManager()->mainWindow()->newWindow();
-        mainWindow->viewManager()->openViewForDoc(doc);
-
-        // use single shot as this action can trigger deletion of this viewspace!
-        QTimer::singleShot(0, this, [this, idx]() {
-            closeTabRequest(idx);
-        });
+        detachDocument(doc);
     }
 }
 
 void KateViewSpace::saveConfig(KConfigBase *config, int myIndex, const QString &viewConfGrp)
 {
     const QString groupname = QString(viewConfGrp + QStringLiteral("-ViewSpace %1")).arg(myIndex);
+    // Ensure that we store the group name. It is used to save config for views inside
+    // the viewspace. See KateViewSpace::saveViewConfig that is called when a view is closed
+    m_group = groupname;
 
-    // aggregate all registered documents & views in view space (LRU ordered)
+    // aggregate all registered documents & views in view space
     // we need even the documents without tabs to avoid that we later have issues with closing
     // documents and not the right ones being used as replacement if you have a limit on tabs
+    // we first store all stuff without visible tabs and then the tabs in tab order to proper restore
+    // that order
     std::vector<KTextEditor::View *> views;
-    QStringList lruList;
-    const auto docList = m_tabBar->lruSortedDocuments();
-    for (auto doc : docList) {
+    QStringList docList;
+    const auto handleDoc = [this, &views, &docList](auto doc) {
         // we can only store stuff about documents that get an id
         const int sessionId = KateApp::self()->documentManager()->documentInfo(doc)->sessionConfigId;
         if (sessionId < 0) {
-            continue;
+            return;
         }
 
-        lruList << QString::number(sessionId);
+        docList << QString::number(sessionId);
         auto it = m_docToView.find(doc);
         if (it != m_docToView.end()) {
             views.push_back(it->second);
         }
+    };
+
+    const QList<KTextEditor::Document *> lruDocList = m_tabBar->lruSortedDocuments();
+    const QList<DocOrWidget> tabs = m_tabBar->documentList();
+    for (auto doc : lruDocList) {
+        // skip stuff with visible tabs
+        if (std::find(tabs.begin(), tabs.end(), DocOrWidget(doc)) == tabs.end()) {
+            handleDoc(doc);
+        }
+    }
+    for (DocOrWidget docOrWidget : tabs) {
+        // only docs are of interest
+        if (KTextEditor::Document *doc = docOrWidget.doc()) {
+            handleDoc(doc);
+        }
     }
 
     KConfigGroup group(config, groupname);
-    group.writeEntry("Documents", lruList);
+    group.writeEntry("Documents", docList);
     group.writeEntry("Count", static_cast<int>(views.size()));
 
     if (currentView()) {
@@ -1261,7 +1309,7 @@ void KateViewSpace::saveConfig(KConfigBase *config, int myIndex, const QString &
             group.writeEntry(QStringLiteral("View %1").arg(idx), sessionId);
 
             // view config, group: "ViewSpace <n> id"
-            QString vgroup = QStringLiteral("%1 %2").arg(groupname, sessionId);
+            QString vgroup = QStringLiteral("%1 %2").arg(groupname, QString::number(sessionId));
             KConfigGroup viewGroup(config, vgroup);
             view->writeSessionConfig(viewGroup);
         }
@@ -1283,9 +1331,15 @@ void KateViewSpace::restoreConfig(KateViewManager *viewMan, const KConfigBase *c
     // set back bar status to configured variant
     tabBarToggled();
 
-    // restore Document lru list so that all tabs from the last session reappear
-    const QStringList lruList = group.readEntry("Documents", QStringList());
-    for (const auto &idOrUrl : lruList) {
+    // Disable opening tabs to the right of current during session restore
+    // as it can mess up the tab order during session restore and because we
+    // open tabs to right as a result of user action
+    const auto openTabsToRightOfCurrent = m_tabBar->openTabsToRightOfCurrent();
+    m_tabBar->setOpenTabsToRightOfCurrentEnabled(false);
+
+    // restore Document list so that all tabs from the last session reappear
+    const QStringList docList = group.readEntry("Documents", QStringList());
+    for (const auto &idOrUrl : docList) {
         // ignore stuff with no id or url
         if (idOrUrl.isEmpty()) {
             continue;
@@ -1304,6 +1358,24 @@ void KateViewSpace::restoreConfig(KateViewManager *viewMan, const KConfigBase *c
         }
     }
 
+    m_tabBar->setOpenTabsToRightOfCurrentEnabled(openTabsToRightOfCurrent);
+
+    // Fix the tab order. If the number of docs cross the tab limit,
+    // then the first tab after the limit will go to position 0, which
+    // might be incorrect. Below we fix the tab order by traversing the
+    // m_registeredDocuments in reverse order and modifying the tabs
+    // in tab bar.
+    // docList: 0 1 2 3
+    // m_registeredDocuments: 0 1 2 3
+    const QList<DocOrWidget> tabsList = m_tabBar->documentList();
+    if (tabsList.size() < docList.size()) {
+        int firstDocIndex = docList.size() - tabsList.size();
+        auto docIt = m_registeredDocuments.begin() + firstDocIndex;
+        for (int i = 0; i < tabsList.size() && docIt != m_registeredDocuments.end(); i++, ++docIt) {
+            m_tabBar->setTabDocument(i, *docIt);
+        }
+    }
+
     // restore active view properties
     if (const QString idOrUrl = group.readEntry("Active View"); !idOrUrl.isEmpty()) {
         // if id use that, is the new format to allow to differentiate between different untitled documents
